
#include<iostream>
#include<sstream>
#include<string>
#include<vector>
#include<map>

#include "Entity.hh"

using namespace std;

/** Check if the Entity has an attribute with the given name.
 * @param attrName Name of the attribute to check
 * @return True if the entity has this attribute
 */
bool Entity::hasAttributeOfName(const std::string& attrName) const {
	return numberAttrs.hasAttributeOfName(attrName) || stringAttrs.hasAttributeOfName(attrName) ||
	boolAttrs.hasAttributeOfName(attrName) || entityAttrs.hasAttributeOfName(attrName);
}

/** Check if a given attribute is static or not.
 * If the attribute does not exist, gives a warning.
 * @param attrName Name of the attribute to check
 * @return True if the attribute is static. False either if the attribute is not static or does not exist.
 */
bool Entity::isStatic(const string& attrName) const {
	if( !(this->hasAttributeOfName(attrName)) ) {
		cerr << "WARNING: trying to make test isStatic on unknown attribute \"" << attrName;
		cerr << "\", false returned" << endl;
		return false;
	}
	return numberAttrs.isStatic(attrName) || stringAttrs.isStatic(attrName) ||
	boolAttrs.isStatic(attrName) || entityAttrs.isStatic(attrName);
}

/** Check if a given attribute is dynamic or not.
 * If the attribute does not exist, gives a warning.
 * @param attrName Name of the attribute to check
 * @return True if the attribute is dynamic. False either if the attribute is not dynamic or does not exist.
 */
bool Entity::isDynamic(const string& attrName) const {
	if( !(this->hasAttributeOfName(attrName)) ) {
		cerr << "WARNING: trying to make test isDynamic on unknown attribute \"" << attrName;
		cerr << "\", false returned" << endl;
		return false;
	}
	return numberAttrs.isDynamic(attrName) || stringAttrs.isDynamic(attrName) ||
	boolAttrs.isDynamic(attrName) || entityAttrs.isDynamic(attrName);
}

/** Check if a given attribute is an atom attribute or not.
 * If the attribute does not exist, gives a warning.
 * @param attrName Name of the attribute to check
 * @return True if the attribute is an atom. False either if the attribute is not an atom or does not exist.
 */
bool Entity::isAtom(const string& attrName) const {
	if( !(this->hasAttributeOfName(attrName)) ) {
		cerr << "WARNING: trying to make test isAtom on unknown attribute \"" << attrName;
		cerr << "\", false returned" << endl;
		return false;
	}
	return numberAttrs.isAtom(attrName) || stringAttrs.isAtom(attrName) ||
	boolAttrs.isAtom(attrName) || entityAttrs.isAtom(attrName);
}

/** Check if a given attribute is a set attribute or not.
 * If the attribute does not exist, gives a warning.
 * @param attrName Name of the attribute to check
 * @return True if the attribute is a set. False either if the attribute is not a set or does not exist.
 */
bool Entity::isSet(const string& attrName) const {
	if( !(this->hasAttributeOfName(attrName)) ) {
		cerr << "WARNING: trying to make test isVector on unknown attribute \"" << attrName;
		cerr << "\", false returned" << endl;
		return false;
	}
	return numberAttrs.isSet(attrName) || stringAttrs.isSet(attrName) ||
	boolAttrs.isSet(attrName) || entityAttrs.isSet(attrName);
}

/** Get the complete list of attributes.
 * @return Return the list of all attributes
 */
list<string> Entity::getAllAttributeNames() const {
	list<string> attrList;
	numberAttrs.getAllAttributeNames(attrList);
	stringAttrs.getAllAttributeNames(attrList);
	boolAttrs.getAllAttributeNames(attrList);
	entityAttrs.getAllAttributeNames(attrList);
	return attrList;
}

/** Make a string from the Entity.
 * @return String representing the Entity
 */
string Entity::toString() const {
	ostringstream st;
	st << "----- " << name << "(type: " << type << ")" << " -----" << endl;
	if(!numberAttrs.empty()) {
		st << "\t** Number attributes **" << endl;
		numberAttrs.toStringStream(st);
		st << endl;
	}
	if(!stringAttrs.empty()) {
		st << "\t** String attributes **" << endl;
		stringAttrs.toStringStream(st);
		st << endl;
	}
	if(!boolAttrs.empty()) {
		st << "\t** Bool attributes **" << endl;
		boolAttrs.toStringStream(st);
		st << endl;
	}
	if(!entityAttrs.empty()) {
		st << "\t** Entity attributes **" << endl;
		entityAttrs.toStringStream(st);
		st << endl;
	}
	if(this->isUnknown()) st << "\t** Entity is UNKNOWN by the agent **" << endl;
	if(this->isKnown()) st << "\t** Entity is KNOWN by the agent **" << endl;
	if(!unknownAttrs.empty()) {
		st << "\t** UNKNOWN attributes **" << endl;
		for(std::set<string>::iterator it = unknownAttrs.begin(); it != unknownAttrs.end(); it++) 
			st << "\t" << *it << endl;
	}
	if(!knownAttrs.empty()) {
		st << "\t** KNOWN attributes **" << endl;
		for(std::set<string>::iterator it = knownAttrs.begin(); it != knownAttrs.end(); it++) 
			st << "\t" << *it << endl;
	}
	return st.str();
}

/** Set the unknown fact to the given value.
 * @param f Value to set unknown to
 */
void Entity::setUnknown(bool f) {
	this->unknown = f;
}

/** Set the unknown fact to the given value.
 * @param f Value to set unknown to
 */
void Entity::setKnown(bool f) {
	this->known = f;
}

/** Initialise the sets of known and unknown attributes.
 */
void Entity::initAttrsKnownUnknown() {
	this->knownAttrs = set<string>();
	this->unknownAttrs = set<string>();
}

/** Set an attribute as unknown depending of the given value.
 * @param attrName Attribute to set
 * @param f If the attribute is already unknown and f is false, removes the attribute. Otherwise if the attribute is not in the unknown list and f is true thenn adds it to the unknown list.
 */
void Entity::setAttrUnknown(const std::string& attrName, bool f) {
	if( !(this->hasAttributeOfName(attrName)) ) 
		cerr << "WARNING: trying to set unknown on non existing attribute \"" << attrName << endl;
	else {
		std::set<string>::iterator it = unknownAttrs.find(attrName);
		if(it != unknownAttrs.end()) {
			if(!f) unknownAttrs.erase(it);
		}
		else {
			if(f) unknownAttrs.insert(attrName);
		}
	}
}

/** Set an attribute as known depending of the given value.
 * @param attrName Attribute to set
 * @param f If the attribute is already known and f is false, removes the attribute. Otherwise if the attribute is not in the known list and f is true thenn adds it to the unknown list.
 */
void Entity::setAttrKnown(const std::string& attrName, bool f) {
	if( !(this->hasAttributeOfName(attrName)) ) 
		cerr << "WARNING: trying to set known on non existing attribute \"" << attrName << endl;
	else {
		std::set<string>::iterator it = knownAttrs.find(attrName);
		if(it != knownAttrs.end()) {
			if(!f) knownAttrs.erase(it);
		}
		else {
			if(f) knownAttrs.insert(attrName);
		}
	}
}

