/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#if !defined HATP_ENTITY_HH_
#define HATP_ENTITY_HH_

#define EMPTY_ENTITY "empty"

#define AGENT "Agent"

#define DEFAULT_NUMBER 0.0
#define DEFAULT_STRING ""
#define DEFAULT_BOOL false
#define DEFAULT_ENTITY NULL

#include<iostream>
#include<string>
#include<list>
#include<cassert>
#include<typeinfo>
#include<string>
#include<vector>
#include<map>
#include<set>
#include<boost/lexical_cast.hpp>

#include "AttributeSet.hh"

typedef double number;

/** Represents any entity in the HATP domains.
 */
class Entity {

private:
	/** Entity's name */
	std::string name;
	/** Entity's type */
	std::string type;

	/**List of number attributes.
	* The implementation of a number is a double.*/
	AttributeSet<number> numberAttrs;
	/**List of string attributes*/
	AttributeSet<std::string> stringAttrs;
	/**List of bool attributes*/
	AttributeSet<bool> boolAttrs;
	/**List of Entity attributes*/
	AttributeSet<Entity*> entityAttrs;

	/** Store the fact that the Entity is unknown for the agent that owns it.
	*/
	bool unknown;
	/** Store the fact that the Entity is known for the agent that owns it.
	*/
	bool known;

	/** Set of unknown attributes.
	* Unkown is used in some domain to say that the Entity's attributes exists but is unkown by an agent.
	*/
	std::set<std::string> unknownAttrs;

	/** Set of known attributes.
	* Known is used in some domain to say that the Entity's attributes exists and is kown by an agent.
	*/
	std::set<std::string> knownAttrs;

public:
	Entity();
	explicit Entity(std::string const& sName, std::string const& sType,
		std::vector<std::string> const& sNumbers, std::vector<std::string> const& dNumbers,
		std::vector<std::string> const& ssNumbers, std::vector<std::string> const& dsNumbers,
		std::vector<std::string> const& sStrings, std::vector<std::string> const& dStrings,
		std::vector<std::string> const& ssStrings, std::vector<std::string> const& dsStrings,
		std::vector<std::string> const& sBools, std::vector<std::string> const& dBools,
		std::vector<std::string> const& ssBools, std::vector<std::string> const& dsBools,
		std::vector<std::string> const& sEntities, std::vector<std::string> const& dEntitiess,
		std::vector<std::string> const& ssEntities, std::vector<std::string> const& dsEntitiess);
	Entity(Entity const& e);
	std::string const& getName() const;
	std::string const& getType() const;
	bool isUnknown() const;
	bool isKnown() const;
	bool isAttrUnknown(const std::string& attrName) const;
	bool isAttrKnown(const std::string& attrName) const;
	std::set<std::string> const& getUnknownAttrs() const;
	std::set<std::string> const& getKnownAttrs() const;
	void setUnknown(bool f);
	void setKnown(bool f);
	void initAttrsKnownUnknown();
	void setAttrUnknown(const std::string& attrName, bool f);
	void setAttrKnown(const std::string& attrName, bool f);
	AttributeSet<number> const& getNumberAttrs() const;
	AttributeSet<std::string> const& getStringAttrs() const;
	AttributeSet<bool> const& getBoolAttrs() const;
	AttributeSet<Entity*> const& getEntityAttrs() const;
	bool isEmpty() const;
	template<typename T> T& get(std::string const& attr);
	template<typename T> T const& get(std::string const& attr) const;
	bool hasAttributeOfName(std::string const& attrName) const;
	bool isStatic(std::string const& attrName) const;
	bool isDynamic(std::string const& attrName) const;
	bool isAtom(std::string const& attrName) const;
	bool isSet(std::string const& attrName) const;
	std::list<std::string> getAllAttributeNames() const;
	std::string toString() const;
	void updateAttrs(std::map<Entity*, Entity*> const& oldToNewValues);
	void delallDynamicVector();
	template<typename T> bool operator==(T const& other) const;
	template<typename T> bool operator!=(T const& other) const;
	template<typename T> bool operator>>(std::set<T> const& other) const;
	template<typename T> bool operator^=(std::set<T> const& other) const;
	~Entity();
};

/** Constructor: creates an entity with EMPTY_ENTITY as name and type.
 * For every attribute list, the default value is given :
 * - number = 0.0
 * - string = ""
 * - bool = false
 * - Entity = NULL
 */
inline Entity::Entity() : name(EMPTY_ENTITY), type(EMPTY_ENTITY), numberAttrs(DEFAULT_NUMBER),
	stringAttrs(DEFAULT_STRING), boolAttrs(DEFAULT_BOOL), entityAttrs(DEFAULT_ENTITY), unknown(false), known(false)
{}

/** Explicit constructor, initialization of AttributeSets.
 * @param sName Entity's name
 * @param sType Entity's type
 * @param sNumbers Static atom numbers (entity attributes)
 * @param dNumbers Dynamic atom numbers
 * @param ssNumbers Static set numbers
 * @param dsNumbers Dynamic set numbers
 * @param sStrings Static atom strings
 * @param dStrings Dynamic atom strings
 * @param ssStrings Static set strings
 * @param dsStrings Dynamic set strings
 * @param sBools Static atom booleans
 * @param dBools Dynamic atom booleans
 * @param ssBools Static set booleans
 * @param dsBools Dynamic set booleans
 * @param sEntities Static atom entities
 * @param dEntities Dynamic atom entities
 * @param ssEntities Static set entities
 * @param dsEntities Dynamic set entities
 */
inline Entity::Entity(std::string const& sName, std::string const& sType,
	std::vector<std::string> const& sNumbers, std::vector<std::string> const& dNumbers,
	std::vector<std::string> const& ssNumbers, std::vector<std::string> const& dsNumbers,
	std::vector<std::string> const& sStrings, std::vector<std::string> const& dStrings,
	std::vector<std::string> const& ssStrings, std::vector<std::string> const& dsStrings,
	std::vector<std::string> const& sBools, std::vector<std::string> const& dBools,
	std::vector<std::string> const& ssBools, std::vector<std::string> const& dsBools,
	std::vector<std::string> const& sEntities, std::vector<std::string> const& dEntities,
	std::vector<std::string> const& ssEntities, std::vector<std::string> const& dsEntities)
	: name(sName), type(sType), 
	numberAttrs(DEFAULT_NUMBER, sNumbers, dNumbers, ssNumbers, dsNumbers),
	stringAttrs(DEFAULT_STRING, sStrings, dStrings, ssStrings, dsStrings),
	boolAttrs(DEFAULT_BOOL, sBools, dBools, ssBools, dsBools),
	entityAttrs(DEFAULT_ENTITY, sEntities, dEntities, ssEntities, dsEntities),
	unknown(false), known(false)
{}

/** Copy constructor.
 * @param e Entity to copy
 */
inline Entity::Entity(Entity const& e) : name(e.getName()), type(e.getType()),
	numberAttrs(e.getNumberAttrs()), stringAttrs(e.getStringAttrs()),
	boolAttrs(e.getBoolAttrs()), entityAttrs(e.getEntityAttrs()), 
	unknown(e.isUnknown()), known(e.isKnown()), unknownAttrs(e.getUnknownAttrs()), knownAttrs(e.getKnownAttrs())
{}

/** Get the name.
 * @return name
 */
inline std::string const& Entity::getName() const {
	return name;
}

/** Get the type.
 * @return type
 */
inline std::string const& Entity::getType() const {
	return type;
}

/** Tell if the entity is unknown.
 * See Entity::unknown.
 * @return True if the entity is unknown
 */
inline bool Entity::isUnknown() const {
	return unknown;
}

/** Tell if the entity is known.
 * See Entity::known.
 * @return True if the entity is known
 */
inline bool Entity::isKnown() const {
	return known;
}

/** Tell if the attribute is unknown.
 * See Entity::unknownAttrs.
 * @param attrName Attribute to check
 * @return True if entity attribute is unknown
 */
inline bool Entity::isAttrUnknown(const std::string& attrName) const {
	std::set<std::string>::const_iterator it = unknownAttrs.find(attrName);
	return (it != unknownAttrs.end());
}

/** Tell if the attribute is known.
 * See Entity::knownAttrs.
 * @param attrName Attribute to check
 * @return True if entity attribute is known
 */
inline bool Entity::isAttrKnown(const std::string& attrName) const {
	std::set<std::string>::const_iterator it = knownAttrs.find(attrName);
	return (it != knownAttrs.end());
}

/** Get the list of unknown attributes.
 * See Entity::unknownAttrs.
 * @return List of unknown attributes
 */
inline std::set<std::string> const& Entity::getUnknownAttrs() const {
	return unknownAttrs;
}

/** Get the list of known attributes.
 * See Entity::knownAttrs.
 * @return List of known attributes
 */
inline std::set<std::string> const& Entity::getKnownAttrs() const {
	return knownAttrs;
}

/** Get all the number attributes.
 * @return List of number attributes
 */
inline AttributeSet<number> const& Entity::getNumberAttrs() const {
	return numberAttrs;
}

/** Get all the string attributes.
 * @return List of string attributes
 */
inline AttributeSet<std::string> const& Entity::getStringAttrs() const {
	return stringAttrs;
}

/** Get all the boolean attributes.
 * @return List of boolean attributes
 */
inline AttributeSet<bool> const& Entity::getBoolAttrs() const {
	return boolAttrs;
}

/** Get all the entity attributes.
 * @return List of entity attributes
 */
inline AttributeSet<Entity*> const& Entity::getEntityAttrs() const {
	return entityAttrs;
}

/** Check if the entity is declared empty.
 * The entity is empty if the constructor with no argument was called. In this case both name and type are equal to EMPTY_ENTITY
 * @return True is the entity is declared empty
 */
inline bool Entity::isEmpty() const {
	return name == EMPTY_ENTITY && type == EMPTY_ENTITY;
}

/** Get a reference to the value of the attribute named attr.
 * This version is for any type T (other than the expected types) so it gives an error and exit the program with a assertion.
 * @param attr Attribute name to get the reference
 * @return Reference to the value of the attribute
 */
template<typename T> inline T& Entity::get(std::string const& attr) {
	std::cerr << "ERROR when accessing attribute \"" << attr << "\"";
	std::cerr << ", type " << typeid(T).name() << " unknown." << std::endl;
	assert(false);
}

/** Get a const reference to the value of the attribute named attr.
 * This version is for any type T (other than the expected types) so it gives an error and exit the program with a assertion.
 * @param attr Attribute name to get the reference
 * @return Const reference to the value of the attribute
 */
template<typename T> inline T const& Entity::get(std::string const& attr) const {
	std::cerr << "ERROR when accessing attribute \"" << attr << "\"";
	std::cerr << ", type " << typeid(T).name() << " unknown." << std::endl;
	assert(false);
}

/** Get a reference to the value of the number attribute named attr.
 * @param attr Number attribute name to get the reference
 * @return Reference to the value of the attribute
 */
template<> inline number& Entity::get<number>(std::string const& attr) {
	try{ return numberAttrs.getAtom(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return numberAttrs.getDefaultAtomValue();
	}
}

/** Get a reference to the value of the number-set attribute named attr.
 * @param attr Number-set attribute name to get the reference
 * @return Reference to the set of value for the attribute
 */
template<> inline std::set<number>& Entity::get<std::set<number> >(std::string const& attr) {
	try{ return numberAttrs.getSet(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return numberAttrs.getDefaultSetValue();
	}
}

/** Get a const reference to the value of the number attribute named attr.
 * @param attr Number attribute name to get the reference
 * @return Const reference to the value of the attribute
 */
template<> inline number const& Entity::get<number>(std::string const& attr) const {
	try{ return numberAttrs.getAtom(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return numberAttrs.getDefaultAtomValue();
	}
}

/** Get a const reference to the value of the number-set attribute named attr.
 * @param attr Number-set attribute name to get the reference
 * @return Const reference to the set of value for the attribute
 */
template<> inline std::set<number> const& Entity::get<std::set<number> >(std::string const& attr) const {
	try{ return numberAttrs.getSet(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return numberAttrs.getDefaultSetValue();
	}
}

/** Get a reference to the value of the string attribute named attr.
 * @param attr String attribute name to get the reference
 * @return Reference to the value of the attribute
 */
template<> inline std::string& Entity::get<std::string>(std::string const& attr) {
	try{ return stringAttrs.getAtom(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value returned." << std::endl;
		return stringAttrs.getDefaultAtomValue();
	}
}

/** Get a reference to the value of the string-set attribute named attr.
 * @param attr String-set attribute name to get the reference
 * @return Reference to the set of value for the attribute
 */
template<> inline std::set<std::string>& Entity::get<std::set<std::string> >(std::string const& attr) {
	try{ return stringAttrs.getSet(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return stringAttrs.getDefaultSetValue();
	}
}

/** Get a const reference to the value of the string attribute named attr.
 * @param attr String attribute name to get the reference
 * @return Const reference to the value of the attribute
 */
template<> inline std::string const& Entity::get<std::string>(std::string const& attr) const {
	try{ return stringAttrs.getAtom(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return stringAttrs.getDefaultAtomValue();
	}
}

/** Get a const reference to the value of the string-set attribute named attr.
 * @param attr String-set attribute name to get the reference
 * @return Cosnt reference to the set of value for the attribute
 */
template<> inline std::set<std::string> const& Entity::get<std::set<std::string> >(std::string const& attr) const {
	try{ return stringAttrs.getSet(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return stringAttrs.getDefaultSetValue();
	}
}

/** Get a reference to the value of the boolean attribute named attr.
 * @param attr Boolean attribute name to get the reference
 * @return Reference to the value of the attribute
 */
template<> inline bool& Entity::get<bool>(std::string const& attr) {
	try{ return boolAttrs.getAtom(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return boolAttrs.getDefaultAtomValue();
	}
}

/** Get a reference to the value of the boolean-set attribute named attr.
 * @param attr Boolean-set attribute name to get the reference
 * @return Reference to the set of value for the attribute
 */
template<> inline std::set<bool>& Entity::get<std::set<bool> >(std::string const& attr) {
	try{ return boolAttrs.getSet(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return boolAttrs.getDefaultSetValue();
	}
}

/** Get a const reference to the value of the boolean attribute named attr.
 * @param attr Boolean attribute name to get the reference
 * @return Const reference to the value of the attribute
 */
template<> inline bool const& Entity::get<bool>(std::string const& attr) const {
	try{ return boolAttrs.getAtom(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return boolAttrs.getDefaultAtomValue();
	}
}

/** Get a const reference to the value of the boolean-set attribute named attr.
 * @param attr Boolean-set attribute name to get the reference
 * @return Const reference to the set of value for the attribute
 */
template<> inline std::set<bool> const& Entity::get<std::set<bool> >(std::string const& attr) const {
	try{ return boolAttrs.getSet(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return boolAttrs.getDefaultSetValue();
	}
}

/** Get a reference to the value of the entity attribute named attr.
 * @param attr Entity attribute name to get the reference
 * @return Reference to the value of the attribute
 */
template<> inline Entity*& Entity::get<Entity*>(std::string const& attr) {
	try{ return entityAttrs.getAtom(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return entityAttrs.getDefaultAtomValue();
	}
}

/** Get a reference to the value of the entity-set attribute named attr.
 * @param attr Entity-set attribute name to get the reference
 * @return Reference to the set of value for the attribute
 */
template<> inline std::set<Entity*>& Entity::get<std::set<Entity*> >(std::string const& attr) {
	try{ return entityAttrs.getSet(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return entityAttrs.getDefaultSetValue();
	}
}

/** Get a const reference to the value of the entity attribute named attr.
 * @param attr Entity attribute name to get the reference
 * @return Const reference to the value of the attribute
 */
template<> inline Entity* const& Entity::get<Entity*>(std::string const& attr) const {
	try{ return entityAttrs.getAtom(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return entityAttrs.getDefaultAtomValue();
	}
}

/** Get a const reference to the value of the entity-set attribute named attr.
 * @param attr Entity-set attribute name to get the reference
 * @return Const reference to the set of value for the attribute
 */
template<> inline std::set<Entity*> const& Entity::get<std::set<Entity*> >(std::string const& attr) const {
	try{ return entityAttrs.getSet(attr); }
	catch(HATPexception& e) {
		std::cerr << "ERROR when accessing an attribute of " << name << ": " << e.description() << std::endl;
		std::cerr << "default value return instead" << std::endl;
		return entityAttrs.getDefaultSetValue();
	}
}

/** Update the entity's attributes.
 * See AttributeSet::updateAttrs
 * @param oldToNewValues A map contaning as key (first row) the previous values to change, and as content (second row) the new values to replace with.
 */
inline void Entity::updateAttrs(std::map<Entity*, Entity*> const& oldToNewValues) {
	entityAttrs.updateAttrs(oldToNewValues);
}

/** Clear all dynamic attributes (set and atom, of any type).
 */
inline void Entity::delallDynamicVector() {
	numberAttrs.clearDynamicAtomAttrs();
	stringAttrs.clearDynamicAtomAttrs();
	boolAttrs.clearDynamicAtomAttrs();
	entityAttrs.clearDynamicAtomAttrs();
	numberAttrs.clearDynamicSetAttrs();
	stringAttrs.clearDynamicSetAttrs();
	boolAttrs.clearDynamicSetAttrs();
	entityAttrs.clearDynamicSetAttrs();
}

/** Make a lexical cast on "other" to convert it to a string and make a comparison to this->name.
 * @param other Object to compare to
 * @return True if match
 */
template<typename T> inline bool Entity::operator==(T const& other) const {
	try {
		std::string sOther = boost::lexical_cast<std::string>(other);
		return name == sOther;
	}
	catch(boost::bad_lexical_cast &) {
		std::cerr << "ERROR: trying to compare an Entity with a type non-convertible in std::string with ==, ";
		std::cerr << "false returned." << std::endl;
		return false;
	}
}

/** Compare two entities by their name.
 * @param e Entity to compare this to
 * @return True if the two entitie names match
 */
template<> inline bool Entity::operator==(Entity const& e) const {
	return name == e.getName();
}

/** Make a lexical cast on "other" to convert it to a string and compare it with this->name to check if different.
 * @param other Object to compare with
 * @return True if different
 */
template<typename T> inline bool Entity::operator!=(T const& other) const {
	try {
		std::string sOther = boost::lexical_cast<std::string>(other);
		return name != sOther;
	}
	catch(boost::bad_lexical_cast &) {
		std::cerr << "ERROR: trying to compare an Entity with a type non-convertible in std::string with !=, ";
		std::cerr << "false returned." << std::endl;
		return false;
	}
}

/** Compare two entities by their name.
 * @param e Entity to compare this to
 * @return True if the two entitie names are different
 */
template<> inline bool Entity::operator!=(Entity const& e) const {
	return name != e.getName();
}

/** "Is in" operator. Check if this is in the given set.
 * The objects stored are cast lexically and compared with this->name.
 * @param other Set to inspect
 * @return True if this (actually this->name) is in the set
 */
template<typename T> inline bool Entity::operator>>(std::set<T> const& other) const {
	typename std::set<T>::const_iterator it;
	try {
		std::string sOther;
		for(it=other.begin() ; it != other.end() ; ++it) {
			sOther = boost::lexical_cast<std::string>(other);
			if(name == sOther) return true;
		}
		return false;
	}
	catch(boost::bad_lexical_cast &) {
		std::cerr << "ERROR: trying to compare an Entity with a type non-convertible in std::string with >>, ";
		std::cerr << "false returned." << std::endl;
		return false;
	}
}

/** "Is in" operator. Check if this is in the given set.
 * The entities are compared using their name.
 * @param s Set to inspect
 * @return True if this (actually this->name) is in the set
 */
template<> inline bool Entity::operator>>(std::set<Entity*> const& s) const {
	for(std::set<Entity*>::const_iterator it=s.begin() ; it != s.end() ; ++it)
		if(name == (*it)->getName()) return true;
	return false;
}

/** "Is not in" operator. Check if this is in the given set.
 * See Entity::operator>>(std::set<T> const& other) const as it is the opposit.
 * @param other Set to inspect
 * @return True if this->name is not in the set
 */
template<typename T> inline bool Entity::operator^=(std::set<T> const& other) const {
	return !(*this >> other);
}

/** Destructor: empty
 */
inline Entity::~Entity() {}

/** Specialised version of toStringStream for entity attributes.
 * See AttributeSet<T>::toStringStream(std::ostringstream& st)
 * @param st The stream the attributes will be printed in
 */
template<> inline void AttributeSet<Entity*>::toStringStream(std::ostringstream& st) const {
	std::map<std::string, Entity*>::const_iterator it1;
	std::map<std::string, std::set<Entity*> >::const_iterator it2;
	std::set<Entity*>::const_iterator setIT;
	std::set<Entity*> setData;
	std::string sNone("<none>");
	for(it1=staticAtomAttrs.begin() ; it1 != staticAtomAttrs.end() ; ++it1) {
		st << "\t\t" << it1->first + ": ";
		if(it1->second) st << it1->second->getName() << "(" << it1->second << " )";
		else st << sNone;
		st << " (static)" << std::endl;
	}
	for(it1=dynamicAtomAttrs.begin() ; it1 != dynamicAtomAttrs.end() ; ++it1) {
		st << "\t\t" << it1->first + ": ";
		if(it1->second) st << it1->second->getName() << "(" << it1->second << ")";
		else st << sNone;
		st << " (dynamic)" << std::endl;
	}
	for(it2=staticSetAttrs.begin() ; it2 != staticSetAttrs.end() ; ++it2) {
		setData = it2->second;
		st << "\t\t" << it2->first << ": {";
		for(setIT=setData.begin() ; setIT != setData.end() ;) {
			st << (*setIT)->getName() << "(" << *setIT << ")";
			if(++setIT != setData.end())
				st << ", ";
		}
		st << "} (static)" << std::endl;
	}
	for(it2=dynamicSetAttrs.begin() ; it2 != dynamicSetAttrs.end() ; ++it2) {
		setData = it2->second;
		st << "\t\t" << it2->first << ": {";
		for(setIT=setData.begin() ; setIT != setData.end() ;) {
			st << (*setIT)->getName() << "(" << *setIT << ")";
			if(++setIT != setData.end())
				st << ", ";
		}
		st << "} (dynamic)" << std::endl;
	}
}

#endif
