#include<iostream>

#include "LightWorldBase.hh"

using namespace std;

/** Check if an LightEntity exists in the agent's database.
 * @param sName Name of the entity to check
 * @return True if the entity exists
 */
bool LightWorldBase::containEntity(std::string const& sName) const {
	return database.find(sName) != database.end();
}

/** Check if an LightEntity (with a specific type) exists in the agent's database.
 * @param sName Name of the entity to check
 * @param sType The type the entity need to be
 * @return True if the entity exists and is of the given type
 */
bool LightWorldBase::containEntity(std::string const& sName, std::string const& sType) const {
	map<string, LightEntity*>::const_iterator it = database.find(sName);
	if(it != database.end() && it->second->getType() == sType) return true;
	return false;
}

/** Find an LightEntity in the world database using its name and type.
 * @param sName The entity to find
 * @param sType The type of the entity
 * @return A pointer to the entity or NULL if the entity does not exist or is not of the given type
 */
LightEntity* LightWorldBase::find(std::string const& sName, std::string const& sType) {
	map<string, LightEntity*>::iterator it = database.find(sName);
	if(it != database.end() && it->second->getType() == sType) return it->second;
	cerr << "ERROR when trying to access to an unknown Entity \"" << sName << "\" of type \"" << sType;
	cerr << "\" in the WorldBase." << endl << "NULL returned instead" << endl;
	return NULL;
}

/** Find an entity in the world database using its name and type.
 * @param sName The entity to find
 * @param sType The type of the entity
 * @return A const pointer to the entity or NULL if the entity does not exist or is not of the given type
 */
LightEntity const* LightWorldBase::find(std::string const& sName, std::string const& sType) const {
	map<string, LightEntity*>::const_iterator it = database.find(sName);
	if(it != database.end() && it->second->getType() == sType) return it->second;
	cerr << "ERROR when trying to access to an unknown Entity \"" << sName << "\" of type \"" << sType;
	cerr << "\" in the WorldBase." << endl << "NULL returned instead" << endl;
	return NULL;
}

/** Find a set of all entities matching the given type.
 * @param sType The type the entities need to match
 * @return Set of entities matching the given type
 */
vector<LightEntity*> LightWorldBase::findByType(string const& sType) const {
	vector<LightEntity*> vec;
	for(map<string,LightEntity*>::const_iterator it=database.begin() ; it!=database.end() ; ++it)
		if(it->second->getType() == sType) vec.push_back(it->second);
	return vec;
}

/** Update an attribute for an agent's entity using the value from another agent.
 * @param a1 Agent from which to copy the value
 * @param a2 Agent that has the entity to be updated
 * @param e Name of the entity to update
 * @param attr Name of the entity's attribute to update
 */
void LightWorldBase::updateForAgent(std::string a1, std::string a2, std::string e, std::string attr) {
	LightEntity* efa1 = this->findForAgent(a1, e);
	LightEntity* efa2 = this->findForAgent(a2, e);
	if(attr == string("NULL") && efa2 != NULL) {
		LightWorldBaseAgent* lwba = WBAgents[a2];
		lwba->replaceEntity(e, efa1);
	}
	else {
		if(efa1 != NULL && efa2 != NULL) {
			if(efa1->isAtom(attr)) {
				if(efa1->getNumberAttrs().hasAttributeOfName(attr)) efa2->get<number>(attr) = efa1->get<number>(attr);
				if(efa1->getStringAttrs().hasAttributeOfName(attr)) efa2->get<string>(attr) = efa1->get<string>(attr);
				if(efa1->getBoolAttrs().hasAttributeOfName(attr)) efa2->get<bool>(attr) = efa1->get<bool>(attr);
				if(efa1->getEntityAttrs().hasAttributeOfName(attr)) efa2->get<LightEntity*>(attr) = efa1->get<LightEntity*>(attr);
			}
			else {
				if(efa1->getNumberAttrs().hasAttributeOfName(attr)) efa2->get<set<number> >(attr) = efa1->get<set<number> >(attr);
				if(efa1->getStringAttrs().hasAttributeOfName(attr)) efa2->get<set<string> >(attr) = efa1->get<set<string> >(attr);
				if(efa1->getBoolAttrs().hasAttributeOfName(attr)) efa2->get<set<bool> >(attr) = efa1->get<set<bool> >(attr);
				if(efa1->getEntityAttrs().hasAttributeOfName(attr)) efa2->get<set<LightEntity*> >(attr) = efa1->get<set<LightEntity*> >(attr);
			}
			efa2->setAttrKnown(attr, false);
			efa2->setAttrUnknown(attr, false);
		}
		else {
			cerr << "ERROR in UpdateForAgent. Something is missing in DB." << endl;
		}
	}
}

/** Update an attribute for an agent's entity using the value from this agent.
 * @param a2 Agent that has the entity to be updated
 * @param efa1 LightEntity with the attribute to copy
 * @param efa2 LightEntity with the attribute to update
 * @param attr Name of the entity's attribute to update
 */
void LightWorldBase::updateForAgent(std::string a2, LightEntity* efa1, LightEntity* efa2, std::string attr) {
	if(a2 == myselfName) return; // blocage: interdit de modifier myself
	if(attr == string("NULL") && efa2 != NULL) {
		LightWorldBaseAgent* lwba = WBAgents[a2];
		lwba->replaceEntity(efa2->getName(), efa1);
	}
	else {
		if(efa1 != NULL && efa2 != NULL) {
			if(efa1->isAtom(attr)) {
				if(efa1->getNumberAttrs().hasAttributeOfName(attr)) efa2->get<number>(attr) = efa1->get<number>(attr);
				if(efa1->getStringAttrs().hasAttributeOfName(attr)) efa2->get<string>(attr) = efa1->get<string>(attr);
				if(efa1->getBoolAttrs().hasAttributeOfName(attr)) efa2->get<bool>(attr) = efa1->get<bool>(attr);
				if(efa1->getEntityAttrs().hasAttributeOfName(attr)) efa2->get<LightEntity*>(attr) = efa1->get<LightEntity*>(attr);
			}
			else {
				if(efa1->getNumberAttrs().hasAttributeOfName(attr)) efa2->get<set<number> >(attr) = efa1->get<set<number> >(attr);
				if(efa1->getStringAttrs().hasAttributeOfName(attr)) efa2->get<set<string> >(attr) = efa1->get<set<string> >(attr);
				if(efa1->getBoolAttrs().hasAttributeOfName(attr)) efa2->get<set<bool> >(attr) = efa1->get<set<bool> >(attr);
				if(efa1->getEntityAttrs().hasAttributeOfName(attr)) efa2->get<set<LightEntity*> >(attr) = efa1->get<set<LightEntity*> >(attr);
			}
			efa2->setAttrKnown(attr, false);
			efa2->setAttrUnknown(attr, false);
		}
	}
}

/** Create a string representing the LightWorldBase.
 * First put the list of entities in this database, then put every agent database. (See LightWorldBaseAgent::toString().)
 * @return A string representing the world entity databse
 */
string LightWorldBase::toString() const {
	string s = "***** Start Myself WorldBase Printing *****\n";
	for(map<string,LightEntity*>::const_iterator it=database.begin() ; it!=database.end() ; ++it)
		s += it->second->toString() + "\n";
	s += "***** End Myself WorldBase Printing *****\n";
	for(map<string,LightWorldBaseAgent*>::const_iterator it2=WBAgents.begin() ; it2!=WBAgents.end() ; ++it2) {
		s += "\nFOR AGENT "+it2->first+" :\n\n";
		s += it2->second->toString() + "\n";
	}
	return s;
}

/** Return a reference to a LightEntity from the agent's database.
 * If the entity does not exist, an empty entity is return instead and a warning is generated.
 * @param c LightEntity to return
 * @return Reference to the entity, or to an empty entity
 */
LightEntity& LightWorldBase::operator[](const char* c) {
	std::map<std::string, LightEntity*>::iterator it = database.find(c);
	if(it != database.end()) return *(it->second);
	else {
		cerr << "ERROR when trying to access to an Entity in the WorldBase: " << c;
		cerr << " has not been found" << endl;
		cerr << "default Entity returned instead" << endl;
		return defaultLightEntity;
	}
}

/** Return a reference to an LightEntity from the agent's database.
 * If the entity does not exist, an empty entity is return instead and a warning is generated.
 * @param s LightEntity to return
 * @return Reference to the entity, or to an empty entity
 */
LightEntity& LightWorldBase::operator[](std::string const& s) {
	std::map<std::string, LightEntity*>::iterator it = database.find(s);
	if(it != database.end()) return *(it->second);
	else {
		cerr << "ERROR when trying to access to an Entity in the WorldBase: " << s;
		cerr << " has not been found" << endl;
		cerr << "default Entity returned instead" << endl;
		return defaultLightEntity;
	}
}

/** Return a reference to an LightEntity from the agent's database.
 * If the entity does not exist, an empty entity is return instead and a warning is generated.
 * @param c LightEntity to return
 * @return Const reference to the entity, or to an empty entity
 */
LightEntity const& LightWorldBase::operator[](const char* c) const {
	std::map<std::string, LightEntity*>::const_iterator it = database.find(c);
	if(it != database.end()) return *(it->second);
	else {
		cerr << "ERROR when trying to access to an Entity in the WorldBase: " << c;
		cerr << " has not been found" << endl;
		cerr << "default Entity returned instead" << endl;
		return defaultLightEntity;
	}
}

/** Return a reference to an LightEntity from the agent's database.
 * If the entity does not exist, an empty entity is return instead and a warning is generated.
 * @param s LightEntity to return
 * @return Const reference to the entity, or to an empty entity
 */
LightEntity const& LightWorldBase::operator[](std::string const& s) const {
	std::map<std::string, LightEntity*>::const_iterator it = database.find(s);
	if(it != database.end()) return *(it->second);
	else {
		cerr << "ERROR when trying to access to an Entity in the WorldBase: " << s;
		cerr << " has not been found" << endl;
		cerr << "default Entity returned instead" << endl;
		return defaultLightEntity;
	}
}

/** Destructor : delete all Entities in database
 */
LightWorldBase::~LightWorldBase() {
	for(map<string, LightEntity*>::iterator it=database.begin() ; it!=database.end() ; ++it)
		delete it->second;
}
